@@ -23,118 +23,6 @@ cmd.errorcodes = {
    CRASH = 99
 }
 
-local function is_ownership_ok(directory)
-   local me = fs.current_user()
-   for _ = 1,3 do -- try up to grandparent
-      local owner = fs.attributes(directory, "owner")
-      if owner then
-         return owner == me
-      end
-      directory = dir.dir_name(directory)
-   end
-   return false
-end
-
-local function exists(file)
-   local fd = io.open(file, "r")
-   if fd then
-      fd:close()
-      return true
-   end
-   return false
-end
-
-do
-   local function Q(pathname)
-      if pathname:match("^.:") then
-         return pathname:sub(1, 2) .. '"' .. pathname:sub(3) .. '"'
-      end
-      return '"' .. pathname .. '"'
-   end
-
-   local function check_lua_version(lua_exe, luaver)
-      if not exists(lua_exe) then
-         return nil
-      end
-      local lv, err = util.popen_read(Q(lua_exe) .. ' -e "io.write(_VERSION:sub(5))"')
-      if luaver and luaver ~= lv then
-         return nil
-      end
-      local ljv
-      if lv == "5.1" then
-         ljv = util.popen_read(Q(lua_exe) .. ' -e "io.write(tostring(jit and jit.version:sub(8)))"')
-         if ljv == "nil" then
-            ljv = nil
-         end
-      end
-      return lv, ljv
-   end
-
-   local find_lua_bindir
-   do
-      local exe_suffix = (package.config:sub(1, 1) == "\\" and ".exe" or "")
-
-      local function insert_lua_versions(names, luaver)
-         local variants = {
-            "lua" .. luaver .. exe_suffix,
-            "lua" .. luaver:gsub("%.", "") .. exe_suffix,
-            "lua-" .. luaver .. exe_suffix,
-            "lua-" .. luaver:gsub("%.", "") .. exe_suffix,
-         }
-         for _, name in ipairs(variants) do
-            names[name] = luaver
-            table.insert(names, name)
-         end
-      end
-
-      find_lua_bindir = function(prefix, luaver)
-         local names = {}
-         if luaver then
-            insert_lua_versions(names, luaver)
-         else
-            for v in util.lua_versions("descending") do
-               insert_lua_versions(names, v)
-            end
-         end
-         if luaver == "5.1" or not luaver then
-            table.insert(names, "luajit" .. exe_suffix)
-         end
-         table.insert(names, "lua" .. exe_suffix)
-
-         local bindirs = { prefix .. "/bin", prefix }
-         local tried = {}
-         for _, d in ipairs(bindirs) do
-            for _, name in ipairs(names) do
-               local lua_exe = dir.path(d, name)
-               table.insert(tried, lua_exe)
-               local lv, ljv = check_lua_version(lua_exe, luaver)
-               if lv then
-                  return name, d, lv, ljv
-               end
-            end
-         end
-         return nil, "Lua interpreter not found at " .. prefix .. "\n" ..
-                     "Tried:\t" .. table.concat(tried, "\n\t")
-      end
-   end
-
-   function cmd.find_lua(prefix, luaver)
-      local lua_interpreter, bindir, luajitver
-      lua_interpreter, bindir, luaver, luajitver = find_lua_bindir(prefix, luaver)
-      if not lua_interpreter then
-         return nil, bindir
-      end
-
-      return {
-         lua_version = luaver,
-         luajit_version = luajitver,
-         lua_interpreter = lua_interpreter,
-         lua_dir = prefix,
-         lua_bindir = bindir,
-      }
-   end
-end
-
 local function check_popen()
    local popen_ok, popen_result = pcall(io.popen, "")
    if popen_ok then
@@ -148,6 +36,19 @@ local function check_popen()
    end
 end
 
+local function check_if_config_is_present(detected, try)
+   local versions = detected.lua_version 
+                    and { detected.lua_version }
+                    or  util.lua_versions("descending")
+   return fun.find(versions, function(v)
+      if util.exists(dir.path(try, ".luarocks", "config-"..v..".lua")) then
+         detected.project_dir = try
+         detected.lua_version = v
+         return detected
+      end
+   end)
+end
+
 local process_tree_flags
 do
    local function replace_tree(flags, root, tree)
@@ -156,20 +57,6 @@ do
       path.use_tree(tree or root)
    end
 
-   local function find_project_dir()
-      local try = "."
-      for _ = 1, 10 do -- FIXME detect when root dir was hit instead
-         local abs = fs.absolute_name(try)
-         if fs.is_dir(abs .. "/.luarocks") and fs.is_dir(abs .. "/lua_modules") then
-            abs = abs:gsub("/.$", "")
-            return abs, abs .. "/lua_modules"
-         elseif fs.exists(abs .. "/.luarocks-no-project") then
-            return nil
-         end
-         try = try .. "/.."
-      end
-   end
-
    local function strip_trailing_slashes()
       if type(cfg.root_dir) == "string" then
         cfg.root_dir = cfg.root_dir:gsub("/+$", "")
@@ -182,10 +69,10 @@ do
       cfg.deploy_lib_dir = cfg.deploy_lib_dir:gsub("/+$", "")
    end
 
-   process_tree_flags = function(flags)
-
-      if cfg.local_by_default then
-         flags["local"] = true
+   process_tree_flags = function(flags, project_dir)
+   
+      if flags["global"] then
+         cfg.local_by_default = false
       end
 
       if flags["tree"] then
@@ -204,28 +91,31 @@ do
             local root_dir = fs.absolute_name(flags["tree"])
             replace_tree(flags, root_dir)
          end
-      elseif flags["project-tree"] then
-         local tree = flags["project-tree"]
-         table.insert(cfg.rocks_trees, 1, { name = "project", root = tree } )
-         loader.load_rocks_trees()
-         path.use_tree(tree)
       elseif flags["local"] then
          if not cfg.home_tree then
             return nil, "The --local flag is meant for operating in a user's home directory.\n"..
                "You are running as a superuser, which is intended for system-wide operation.\n"..
                "To force using the superuser's home, use --tree explicitly."
-         end
-         replace_tree(flags, cfg.home_tree)
-      else
-         local project_dir, rocks_tree = find_project_dir()
-         if project_dir then
-            table.insert(cfg.rocks_trees, 1, { name = "project", root = rocks_tree } )
-            loader.load_rocks_trees()
-            path.use_tree(rocks_tree)
          else
-            local trees = cfg.rocks_trees
-            path.use_tree(trees[#trees])
+            replace_tree(flags, cfg.home_tree)
          end
+      elseif flags["project-tree"] then
+         local tree = flags["project-tree"]
+         table.insert(cfg.rocks_trees, 1, { name = "project", root = tree } )
+         loader.load_rocks_trees()
+         path.use_tree(tree)
+      elseif cfg.local_by_default then
+         if cfg.home_tree then
+            replace_tree(flags, cfg.home_tree)
+         end
+      elseif project_dir then
+         local project_tree = project_dir .. "/lua_modules"
+         table.insert(cfg.rocks_trees, 1, { name = "project", root = project_tree } )
+         loader.load_rocks_trees()
+         path.use_tree(project_tree)
+      else
+         local trees = cfg.rocks_trees
+         path.use_tree(trees[#trees])
       end
 
       strip_trailing_slashes()
@@ -262,6 +152,33 @@ local function process_server_flags(flags)
    return true
 end
 
+local function get_lua_version(flags)
+   if flags["lua-version"] then
+      return flags["lua-version"] 
+   end
+   local dirs = {
+      dir.path(cfg.home_tree, ".luarocks"),
+      cfg.sysconfdir,
+   }
+   if flags["project-tree"] then
+      table.insert(dirs, 1, dir.path(flags["project-tree"], "..", ".luarocks"))
+   end
+   for _, d in ipairs(dirs) do
+      local f = dir.path(d, "default-lua-version.lua")
+      local mod, err = loadfile(f, "t")
+      if mod then
+         local pok, ver = pcall(mod)
+         if pok and type(ver) == "string" and ver:match("%d+.%d+") then
+            if flags["verbose"] then
+               util.printout("Defaulting to Lua " .. ver .. " based on " .. f .. " ...")
+            end
+            return ver
+         end
+      end
+   end
+   return nil
+end
+
 --- Main command-line processor.
 -- Parses input arguments and calls the appropriate driver function
 -- to execute the action requested on the command-line, forwarding
@@ -359,49 +276,77 @@ function cmd.run_command(description, commands, external_namespace, ...)
    end
    command = command:gsub("-", "_")
 
+   if command == "config" then
+      if nonflags[1] == "lua_version" and nonflags[2] then
+         flags["lua-version"] = nonflags[2]
+      elseif nonflags[1] == "lua_dir" and nonflags[2] then
+         flags["lua-dir"] = nonflags[2]
+      end
+   end
+
+   local lua_version = get_lua_version(flags)
+
    if flags["deps-mode"] and not deps.check_deps_mode_flag(flags["deps-mode"]) then
       die("Invalid entry for --deps-mode.")
    end
+   
+   local lua_found = false
 
-   local project_dir
-   if flags["project-tree"] then
-      project_dir = flags["project-tree"]:gsub("[/\\][^/\\]+$", "")
-   end
-
-   local lua_data
+   local detected
    if flags["lua-dir"] then
       local err
-      lua_data, err = cmd.find_lua(flags["lua-dir"], flags["lua-version"])
-      if not lua_data then
+      detected, err = util.find_lua(flags["lua-dir"], lua_version)
+      if not detected then
          die(err)
       end
-   elseif flags["lua-version"] then
+      lua_found = true
+      assert(detected.lua_version)
+      assert(detected.lua_dir)
+   elseif lua_version then
       local path_sep = (package.config:sub(1, 1) == "\\" and ";" or ":")
       for bindir in os.getenv("PATH"):gmatch("[^"..path_sep.."]+") do
          local parentdir = bindir:gsub("[\\/][^\\/]+[\\/]?$", "")
-         lua_data = cmd.find_lua(dir.path(parentdir), flags["lua-version"])
-         if lua_data then
+         detected = util.find_lua(dir.path(parentdir), lua_version)
+         if detected then
             break
          end
-         lua_data = cmd.find_lua(bindir, flags["lua-version"])
-         if lua_data then
+         detected = util.find_lua(bindir, lua_version)
+         if detected then
             break
          end
       end
-      if not lua_data then
-         die("Could not find a Lua interpreter for version " .. flags["lua-version"] .. " in your PATH")
+      if detected then
+         lua_found = true
+      else
+         detected = {
+            lua_version = lua_version,
+         }
       end
-   elseif project_dir then
-      local lua_version
-      for v in util.lua_versions("descending") do
-         if exists(dir.path(project_dir, ".luarocks", "config-"..v..".lua")) then
-            lua_version = v
+   end
+
+   if flags["project-tree"] then
+      local project_tree = flags["project-tree"]:gsub("[/\\][^/\\]+$", "")
+      detected = detected or {}
+      detected.project_dir = project_tree
+      local d = check_if_config_is_present(detected, project_tree)
+      if d then
+         detected = d
+      end
+   else
+      detected = detected or {}
+      local try = "."
+      for _ = 1, 10 do -- FIXME detect when root dir was hit instead
+         if util.exists(try .. "/.luarocks") and util.exists(try .. "/lua_modules") then
+            local d = check_if_config_is_present(detected, try)
+            if d then
+               detected = d
+               break
+            end
+         elseif util.exists(try .. "/.luarocks-no-project") then
             break
          end
+         try = try .. "/.."
       end
-      lua_data = {
-         lua_version = lua_version
-      }
    end
 
    -- FIXME A quick hack for the experimental Windows build
@@ -418,7 +363,7 @@ function cmd.run_command(description, commands, external_namespace, ...)
    end
 
    -----------------------------------------------------------------------------
-   local ok, err = cfg.init(lua_data, project_dir, util.warning)
+   local ok, err = cfg.init(detected, util.warning)
    if not ok then
       die(err)
    end
@@ -426,6 +371,28 @@ function cmd.run_command(description, commands, external_namespace, ...)
 
    fs.init()
 
+   if not lua_found then
+      if cfg.variables.LUA_DIR then
+         local found = util.find_lua(cfg.variables.LUA_DIR, cfg.lua_version)
+         if found then
+            lua_found = true
+         end
+      end
+   end
+
+   if not lua_found then
+      util.warning("Could not find a Lua interpreter for version " ..
+                   lua_version .. " in your PATH. " ..
+                   "Modules may not install with the correct configurations. " ..
+                   "You may want to specify to the path prefix to your build " ..
+                   "of Lua " .. lua_version .. " using --lua-dir")
+   end
+   cfg.lua_found = lua_found
+
+   if detected.project_dir then
+      detected.project_dir = fs.absolute_name(detected.project_dir)
+   end
+
    if flags["version"] then
       util.printout(program.." "..cfg.program_version)
       util.printout(description)
@@ -448,7 +415,7 @@ function cmd.run_command(description, commands, external_namespace, ...)
       die("Current directory does not exist. Please run LuaRocks from an existing directory.")
    end
 
-   ok, err = process_tree_flags(flags)
+   ok, err = process_tree_flags(flags, detected.project_dir)
    if not ok then
       die(err)
    end
@@ -468,15 +435,9 @@ function cmd.run_command(description, commands, external_namespace, ...)
       end
    end
 
-   if not is_ownership_ok(cfg.local_cache) then
-      util.warning("The directory '" .. cfg.local_cache .. "' or its parent directory "..
-                   "is not owned by the current user and the cache has been disabled. "..
-                   "Please check the permissions and owner of that directory. "..
-                   (cfg.is_platform("unix")
-                    and ("If executing "..util.this_program("luarocks").." with sudo, you may want sudo's -H flag.")
-                    or ""))
-      cfg.local_cache = fs.make_temp_dir("local_cache")
-      util.schedule_function(fs.delete, cfg.local_cache)
+   -- if running as superuser, use system cache dir
+   if not cfg.home_tree then
+      cfg.local_cache = dir.path(fs.system_cache_dir(), "luarocks")
    end
 
    if commands[command] then
